package org.ylhealth.ym.springtest.hibernate;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.TimeUnit;
import javax.persistence.Id;
import org.hibernate.SQLQuery;
import org.hibernate.StatelessSession;
import org.hibernate.engine.spi.SessionFactoryImplementor;
import org.hibernate.event.spi.PostLoadEvent;
import org.hibernate.event.spi.PostLoadEventListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.i18n.LocaleContextHolder;
import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;

/** */
public class I18NReadEventListener implements PostLoadEventListener {
  /**
   * 
   */
  private static final long serialVersionUID = -7374424944903271183L;

  private Logger logger = LoggerFactory.getLogger(getClass());

  private SessionFactoryImplementor sessionFactory;
  @SuppressWarnings("rawtypes")
  private Cache<String, List> cache =
      CacheBuilder.newBuilder().expireAfterAccess(30, TimeUnit.MINUTES).build();

  public I18NReadEventListener(SessionFactoryImplementor sessionFactory) {
    this.sessionFactory = sessionFactory;
  }

  @SuppressWarnings("rawtypes")
  @Override
  public void onPostLoad(PostLoadEvent event) {
    Object entity = event.getEntity();
    Class clazz = entity.getClass();
    Collection<Field> fields = getLocalizedFields(clazz);
    if (!fields.isEmpty())
      try (StatelessSession session =
          sessionFactory.openStatelessSession(event.getSession().connection())) {

        Object id = getPKValue(clazz, entity);
        
        String language = LocaleContextHolder.getLocale().toString();

        for (Field field : fields) {
          List result = getTranslate(session, id, field, language);
          if (!result.isEmpty()) field.set(entity, result.get(0));
        }
      } catch (Exception e) {
        logger.error("ReadEventListener error", e);
        throw new IllegalStateException();
      }
  }

  private List getTranslate(StatelessSession session, Object id, Field field, String language) {
    String key = field + "_" + id + "_" + language;
    List result = cache.getIfPresent(key);
    if (result == null) {
      I18NTranslate localized = field.getAnnotation(I18NTranslate.class);
      String sql =
          String.format(
              "select %s from %s where %s = ? and languageCode = ?",
              localized.column(), localized.table(), localized.id());
      SQLQuery query = session.createSQLQuery(sql);
      query.setParameter(0, id);
      query.setParameter(1, language);
      result = query.list();
      logger.info("SQL: {}: {}", sql, result);
      cache.put(key, result);
    }
    return result;
  }

  public static Object getPKValue(Class<?> clazz, Object entity)
      throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
    Collection<Field> fields = getAllDeclaredFields(clazz);
    for (Field field : fields) {
      if (field.getAnnotation(Id.class) != null) {
        field.setAccessible(true);
        return field.get(entity);
      }
    }
    for (Method method : clazz.getDeclaredMethods()) {
      if (method.getAnnotation(Id.class) != null) {
        return method.invoke(entity, null);
      }
    }
    return null;
  }
  /**
   * Returns the entity's @{@link I18NTranslate} fields.
   *
   * <p>These fields are made accessible.
   */
  private static Collection<Field> getLocalizedFields(Class<?> clazz) {
    Collection<Field> fields = getAllDeclaredFields(clazz);
    List<Field> localizedFields = new ArrayList<>();
    fields
        .stream()
        .filter(field -> field.getAnnotation(I18NTranslate.class) != null)
        .forEach(
            field -> {
              field.setAccessible(true);
              localizedFields.add(field);
            });
    return localizedFields;
  }

  private static Collection<Field> getAllDeclaredFields(Class<?> clazz) {
    List<Field> fields = new ArrayList<>();
    Collections.addAll(fields, clazz.getDeclaredFields());
    if (clazz.getSuperclass() != null) {
      fields.addAll(getAllDeclaredFields(clazz.getSuperclass()));
    }
    return fields;
  }
}
